#ifndef BASICSOLVER_HPP_
#define BASICSOLVER_HPP_

#include    <vector>

#include    "ConstraintDB.hpp"
#include    "Types.hpp"
#include    "Settings.hpp"
#include    "Statistics.hpp"
#include    "DecayHeap.hpp"
#include    "BitVector.hpp"
#include 	"ConflictsRestarts.hpp"
#include 	"Simplifier.hpp"

using        std::vector;

enum BackjumpType{BACKJUMP_SET_PREV,
	BACKJUMP_SET_PREV_FLP,
	//BACKJUMP_SET_PREV_RAND,
	BACKJUMP_PREV_UNTOUCHED};


enum OnConflict{RETURN,INIT,UNLINK};

class BasicSolver: public ConstraintDB{

protected:
	static const int32_t BCP_SUCCESS = -1; //SOMETHING < 0

    /** A heuristic measurement of the activity of a variable  */
    DecayHeap< Variable, var_act_t >   var_act_hp;

    /** Indicates if state of the solver is ok*/
    bool                            state_ok;

    /** Assignment stack; stores all assigments made in the order they were made*/
    vector<Literal>                 assign_stack;

    /** Head of queue - as index into the assign_stack */
    uint32_t                        queue_begin;

    /** Indices for different decision levels in 'assign_stack' */
    vector<uint32_t>                decision_level_end;

    /** Number of top-level assignments at last execution of 'simplify()' */
    uint32_t                        simplify_assign_idx;

    /** Remaining number of propagations that must be made before next execution of 'simplify()' */
    int64_t                         simplify_props_remaining;

    /** Data structure to set a flag for a variable and allow efficient clearing of all flags */
    BitVector			            var_flags;

    uint32_t						clause_stamp;

    Conflicts2MaxLearnts			conflict_count;
    LubyRestarts					luby_restarts;

    //conflict related information
    vector<Literal>					conflict_lemma;
    uint64_t						conflict_involved_levels;
    vector<Variable>                var_flags_to_clear;
    vector<Literal>                 tmp_literal_store;

    //for simplifier:
    Simplifier*						sapplifyalot;

//internal Methods

    inline uint32_t numberAssignments ()  const;
    inline void		reinsertVarOrder    (Variable v);
    inline void     newDecisionLevel    ();
    template<class Reason>
    inline void     uncheckedEnqueue    (const Literal& lit, Reason from = (Clause*)NULL);
    inline void     varDecayActivity    ();
    inline void     claDecayActivity    ();
    inline uint32_t decisionLevel       ()  const;

    void 			clearBinWatchers(Variable var);
    void 			clearBinWatchers(Literal lit);
    void			simplifyAllBinWatchers ();
    void			simplifyBinWatchersOf (vector<WatchedElement>& wes);
    TrBool          simplify            ( );


    void			firstDecisionCheckBothPol(Literal next);
    int32_t        	bcp           		(const OnConflict on_conflict );
    int32_t        	bcpPropagateLit		(const OnConflict on_conflict, const Literal& false_lit);
    inline void 	bcpFinishOnConflict	(we_vec&  watched, we_iter iter, we_iter copy,we_iter end);
    inline int32_t	bcpInitConflict		(const int32_t& dec_level, const Literal& w1, const Literal& w2);
    inline bool		bcpFindNewWatcher	(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx);

    inline void    	backjumpKeepLevel   (uint32_t level,const BackjumpType type);    // Backtrack until a certain level.
    void    		backjumpUndoAssigns (uint32_t assign_idx, const BackjumpType type); //undoes all assignments above and including the specified assign_idx
    bool        	firstUIP            (int32_t nbr_pivot_lits);
    inline void     firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int32_t& nbr_pivot_elements);
    uint32_t        backtrackLevel      ();
    void            reduceSelfImpl  	();
    inline void		unsetConfLemmaFlags	();
    bool            litIsNecessary      (Literal lit);
    inline bool     reduceSelfImpCheckLiteral (const Literal& lit);

    inline void		varBump				(const Variable& var);
    TrBool          search             	(const OnConflict on_conflict,uint64_t abs_max_confl );
    TrBool  		solveConsiderTimeAndStarts( const double timeout,int32_t starts);
    Literal         nextDecision();
    bool            uncheckedAddClause  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend);

    //for simplifier:
    Clause*         uncheckedAddClauseForSimplifier  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend);

    uint32_t		currentId;

    struct LitActDescendingFunctor{
    	BasicSolver* outer ;
    	LitActDescendingFunctor(BasicSolver* s):outer(s){};
      	bool operator() (const Literal& a, const Literal& b){return outer->var_act_hp.getValue(a.toVar()) > outer->var_act_hp.getValue(b.toVar());	}
    };


public:
                    BasicSolver(Settings& sets, Statistics& stats);
                    ~BasicSolver();

    inline  void    setInvalid();
    inline  void    reserveForVariables(uint32_t sz);
    inline  void    maybeNewVariable(Variable v);
            bool    addClause(vector<Literal>& lits);
    uint64_t		conflictCount() const {return conflict_count.getConflicts();}

            TrBool  isSAT(uint32_t timeout = ~zero32);
            void	takePartialModelAsUnits();
            TrBool	extendModel();

            //for simplifyer
			bool	addClauseFromSimplifier(Clause* myClause);
			void	reserveForClausesSimplifier (uint32_t need_sz);
			bool    addClauseForSimplifier(vector<Literal>& lits);

};



//Inline Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void    BasicSolver::setInvalid()    {state_ok = false;}

void    BasicSolver::reinsertVarOrder(Variable v){
	if (!var_act_hp.contains(v)) var_act_hp.restoreElement(v);
}


void		BasicSolver::varBump(const Variable& var){
	if(!var_act_hp          .bump(var)) ++statistics.nbr_expensive_var_bmp;
}

template<class Reason>
void    BasicSolver::uncheckedEnqueue(const Literal& lit, Reason from){
    asrt(litIsUnknown(lit));
    varUncheckedAssign(lit,from,decisionLevel());
    assign_stack.push_back(lit);
}


void        BasicSolver::claDecayActivity()            		{cls_inc *= settings.cls_decay; }

uint32_t    BasicSolver::decisionLevel ()			const	{return decision_level_end.size(); }

uint32_t	BasicSolver::numberAssignments()		const	{return assign_stack.size();}

void        BasicSolver::newDecisionLevel()                	{decision_level_end.push_back(assign_stack.size()); }



void     	BasicSolver::reserveForVariables(uint32_t sz){
	if(sz > numberVariables()){
		ConstraintDB::reserveForVariables(sz);
		var_act_hp.reserve(sz);

		if(settings.simp)
			(*sapplifyalot).reserveSpaceForVars(sz);
	}
}


void 		BasicSolver::maybeNewVariable(Variable v){
    reserveForVariables(v.idx() + 1);

    if(!var_act_hp.contains(v)){ var_act_hp.lazyAddNewElement(v,var_act_t(0));}
}


void    	BasicSolver::backjumpKeepLevel(uint32_t level,BackjumpType type){

    if(decisionLevel() <= level) return;
    backjumpUndoAssigns(decision_level_end[level],type);

    decision_level_end    	.resize(level);
}


void BasicSolver::bcpFinishOnConflict(we_vec&  watched, we_iter iter, we_iter copy,we_iter end){

	queue_begin 	= assign_stack.size(); // to finish outer while loop
	while(iter != end) *copy++ = *iter++;
	if(copy != end) watched.erase(copy,end);
}


int32_t	BasicSolver::bcpInitConflict(const int32_t& dec_level, const Literal& w1, const Literal& w2){
	asrt(var_flags.allCleared());

	conflict_lemma					.resize(1,INVALID_LIT);
	conflict_involved_levels 		= 0;
	int32_t nbr_pivot_lits 			= 0;
	firstUIPCheckLiteral(w1,dec_level,nbr_pivot_lits);
	firstUIPCheckLiteral(w2,dec_level,nbr_pivot_lits);
	return nbr_pivot_lits;
}


bool	BasicSolver::bcpFindNewWatcher(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx){
	uint32_t this_lit_idx = other_lit_idx^1;

	for(uint32_t k=cl.getNextK(clause_stamp);k<cl.size();++k){
	//for(uint32_t k=2;k<cl.size();++k){
		if(litIsNotFalse(cl[k])){
	    	cl[this_lit_idx] = cl[k];
	        cl[k]            = false_lit;
	        cl.setNextK(k+1,clause_stamp);

	        //we is "deleted" (not copied) from the watchers list
	        litAddWatchedClause(cl[this_lit_idx],cl[other_lit_idx],&cl);
	        return true;
		}
	}
	return false;
}


void BasicSolver::firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int& nbr_pivot_elements){
   const Variable      var(lit.toVar());

   if(var_flags.isSet(var.idx())) return;
   asrt(litIsFalse(lit));

   int level = varAssignLevel(var);
   if(level > 0){
       var_flags.setFlag(var.idx());
       varBump(var);

       if(level <  dec_level){
           conflict_lemma			.push_back(lit);
           conflict_involved_levels |= hash64(level);
       }else {
           ++nbr_pivot_elements;
       }
   }
}

void BasicSolver::unsetConfLemmaFlags(){
	uint32_t cp = conflict_lemma.size();
	while(cp--) {var_flags.carelessClear(conflict_lemma[cp].toVar().idx());}
	for(cp = var_flags_to_clear.size();cp--;) {var_flags.carelessClear(var_flags_to_clear[cp].idx());}
	var_flags_to_clear.clear();
	asrt(var_flags.allCleared());
}


bool BasicSolver::reduceSelfImpCheckLiteral(const Literal& lit){

    Variable var = lit.toVar();

    if(var_flags.isSet(var.idx())) return false;

    const AssignInfo& a_info = varAssignInfo(var);
    if(a_info.getLevel() > 0){
        if(a_info.noReason() || ((hash64(a_info.getLevel()) & conflict_involved_levels) == 0)){
            return true;
        }

        var_flags.setFlag(var.idx());
        tmp_literal_store.push_back(lit);
    }
    return false;
}




#define for_all_but_one(idx,sz,todo) {todo}; for(idx = 2;idx < sz;++idx){ todo}


const char*  basic_solver_cpp_version();
const char*  sapperlot_cpp_version();
#define basic_solver_hpp_version "BasicSolver.hpp:    $Revision: 1.1 $"


























#endif /*BASICSOLVER_HPP_*/
